@@ -564,18 +564,7 @@ const char *
 cygheap_user::env_systemroot (const char *name, size_t namelen)
 {
   if (!psystemroot)
-    {
-      int size = GetSystemWindowsDirectoryW (NULL, 0);
-      if (size > 0)
-	{
-	  WCHAR wsystemroot[size];
-	  size = GetSystemWindowsDirectoryW (wsystemroot, size);
-	  if (size > 0)
-	    sys_wcstombs_alloc (&psystemroot, HEAP_STR, wsystemroot);
-	}
-      if (size <= 0)
-	debug_printf ("GetSystemWindowsDirectoryW(), %E");
-    }
+    sys_wcstombs_alloc (&psystemroot, HEAP_STR, windows_directory);
   return psystemroot;
 }
 
@@ -1743,7 +1732,7 @@ pwdgrp::fetch_account_from_file (fetch_user_arg_t &arg)
       /* Override SID with SID string. */
       arg.sid->string (str);
       arg.name = str;
-      /*FALLTHRU*/
+      fallthrough;
     case NAME_arg:
       arg.len = strlen (arg.name);
       break;
@@ -2300,7 +2289,7 @@ pwdgrp::fetch_account_from_windows (fetch_user_arg_t &arg, cyg_ldap *pldap)
 	      if (!its_ok)
 		return NULL;
 	    }
-	  /*FALLTHRU*/
+	  fallthrough;
 	case SidTypeGroup:
 	case SidTypeAlias:
 	  /* Predefined alias? */
